from collections.abc import Callable

from acte.build.type import Prop, to_ref
from acte.build.viewer.common.base import Base
from acte.node import Text
from acte.state import Compute, Ref


class TextViewer(Base):
    @classmethod
    def text(cls, content: Callable[[], str] | Prop[str]) -> None:
        cls._check_skip()

        if callable(content):
            content = Compute(content)

        cls._append_awaitable(cls._text_constructor(content))

    @classmethod
    async def _text_constructor(cls, content: Prop[str]) -> None:
        content = to_ref(content)

        node = Text()
        await node.bind_content(content)

        cls._attach_to_container(node)



from typing import TypeVar, TypeAlias

from acte.state import Ref

T = TypeVar('T')

Prop: TypeAlias = Ref[T] | T


def to_ref(value: Prop[T] | None) -> Ref[T]:
    if isinstance(value, Ref):
        return value

    return Ref(value)


as_prop = to_ref



from .build import Viewer, Component, as_prop, Prop

from . import state
from . import chatbot
from . import server
from . import session

v = Viewer

__all__ = [
    "v",
    "as_prop",
    "Prop",
    "Component",
    "state",
    "chatbot",
    "server",
    "session",
    "schema",
]



from __future__ import annotations

from typing import TypeVar, TYPE_CHECKING

from acte.context.context import Context
from acte.state.ref import Ref

if TYPE_CHECKING:
    from acte.state.effect import Effect

T = TypeVar('T')


class Signal(Ref[T]):
    def __init__(self, value: T | None) -> None:
        super().__init__(value)

        self._effect_list: list[Effect] = []

    @property
    def value(self) -> T | None:
        effect_stack = Context.get_effect_stack()

        # if no effect is running, return the value directly
        if len(effect_stack) == 0:
            return self._value

        # add this signal to the current effect
        current_effect = effect_stack[-1]
        if current_effect not in self._effect_list:
            # signal and effect add each other
            self.add_effect(current_effect)
            current_effect.add_signal(self)

        return self._value

    async def set(self, value: T | None) -> None:
        if self._value == value:
            return

        self._value = value

        prev_effect_list = self._effect_list

        # remove all effects and related signal
        self._effect_list = []
        for effect in prev_effect_list:
            effect.remove_signal(self)

        for effect in prev_effect_list:
            if effect.is_cancel is False:  # parent effect change may have cancelled child effect by its own
                await effect.call()

    def add_effect(self, effect: Effect) -> None:
        self._effect_list.append(effect)

    def remove_effect(self, effect: Effect) -> None:
        self._effect_list.remove(effect)



import traceback
import uuid
from typing import Type, Any, cast

from acte.build import Builder, Component
from acte.executor import Executor, ActionType
from acte.render import Renderer, HtmlRenderer
from acte.session.exception import SessionBuildException, SessionDisplayException, SessionValidateException, \
    SessionExecuteException

from acte.session.session import Session
from acte.session.ret_val import RetVal
from acte.session.validator import Validator


class SessionManager(Validator):
    def __init__(
            self,
            entry_class: Type[Component],
            builder: Builder | None = None,
            executor: Executor | None = None,
            renderer: Renderer | None = None,
            debug: bool = False,
    ) -> None:
        self._entry_class = entry_class
        self._builder = builder if builder is not None else Builder()
        self._executor = executor if executor is not None else Executor()
        self._renderer = renderer if renderer is not None else HtmlRenderer()
        self._debug = debug

        self._session_dict: dict[str, Session] = {}

    async def new_session(self) -> dict[str, Any]:
        ret_val = RetVal()

        session = Session(self._entry_class, self._builder, self._executor, self._renderer)

        session_id = str(uuid.uuid4())
        self._session_dict[session_id] = session

        ret_val.session_id = session_id

        try:
            await session.start()
        except Exception as e:
            if self._debug:
                traceback.print_exc()

            ret_val.build_error = str(e)
            return ret_val.to_dict()

        screen, display_error = await self._display(session)
        if screen is not None:
            ret_val.screen = screen

        if display_error is not None:
            if self._debug:
                traceback.print_exc()

            ret_val.display_error = display_error

        return ret_val.to_dict()

    async def execute(self, request: dict[str, Any]) -> dict[str, Any]:
        """
        :param request:
        {
            "session_id": str,
            "actions": [
                {
                    "target_id": str,
                    "type": str,
                    "value": str | None
                }
            ]
        }
        """
        ret = RetVal()

        ret.validate_error = self.validate_execute_request(request, self._session_dict)
        if ret.validate_error is not None:
            if self._debug:
                traceback.print_exc()

            sid = request.get('session_id')
            if (sid is not None) and (sid in self._session_dict):
                ret.session_id = sid

            ret.executed_actions = []

            return ret.to_dict()

        ret.session_id = request['session_id']
        actions = request['actions']

        session = self._session_dict[cast(str, ret.session_id)]

        executed_actions = []
        for action in actions:
            target_id = action['target_id']
            action_type = ActionType(action['action_type'])
            value = action.get('value')

            try:
                await session.execute(target_id, action_type, value)
                executed_actions.append(action)
            except Exception as e:
                if self._debug:
                    traceback.print_exc()

                ret.execute_error = {
                    "action": action,
                    "message": str(e)
                }

                ret.executed_actions = executed_actions
                break

        ret.screen, ret.display_error = await self._display(session)

        return ret.to_dict()

    async def display(self, request: dict[str, Any]) -> dict[str, Any]:
        ret = RetVal()

        ret.validate_error = self._validate_session_id(request, self._session_dict)
        if ret.validate_error is not None:
            if self._debug:
                traceback.print_exc()

            sid = request.get('session_id')
            if (sid is not None) and (sid in self._session_dict):
                ret.session_id = sid

            return ret.to_dict()

        sid = request['session_id']
        ret.session_id = sid
        session = self._session_dict[cast(str, sid)]

        ret.screen, ret.display_error = await self._display(session)

        return ret.to_dict()

    async def _display(self, session: Session) -> tuple[str | None, str | None]:
        try:
            await session.display()
            screen = session.render()
            return screen, None
        except Exception as e:
            if self._debug:
                traceback.print_exc()
            return None, str(e)



import json

from acte.build import call_on_display
from acte.common.util import call_mix
from acte.executor.action_type import ActionType
from acte.node import Root, Node, Button, Input, ComponentNode
from acte.node.implement import Interactive, Container


class Executor:
    @classmethod
    async def execute(cls, root: Root, target_id: str, action_type: ActionType, value: str | None) -> None:
        target = cls._find_node(root, target_id)
        if target is None:
            raise ValueError(f"Target not found: {target_id}")

        await cls._execute(target, action_type, value)

    @classmethod
    async def display(cls, node: Node) -> None:
        if isinstance(node, ComponentNode):
            c = node.component
            if c is not None:
                await call_on_display(c)

        if isinstance(node, Container):
            for child in node.children:
                await cls.display(child)

    @classmethod
    def _find_node(cls, node: Node, interactive_id: str) -> Interactive | None:
        if isinstance(node, Interactive) and node.interactive_id == interactive_id:
            return node

        if not isinstance(node, Container):
            return None

        for child in node.children:
            target = cls._find_node(child, interactive_id)
            if target is not None:
                return target

        return None

    @classmethod
    async def _execute(cls, interactive: Interactive, action_type: ActionType, value: str | None) -> None:
        if isinstance(interactive, Button):
            await cls._button_execute(interactive, action_type)
        elif isinstance(interactive, Input):
            await cls._input_execute(interactive, action_type, value)
        else:
            raise ValueError(f"Unknown interactive type: {interactive}")

    @classmethod
    async def _button_execute(cls, button: Button, action_type: ActionType) -> None:
        if action_type == ActionType.PRESS:
            if button.on_press is not None:
                await call_mix(button.on_press)
        else:
            raise ValueError(f"Button doesn't support action type: {action_type.value}")

    @classmethod
    async def _input_execute(cls, node: Input, action_type: ActionType, value: str | None) -> None:
        if action_type == ActionType.SET:
            if value is None:
                raise ValueError("Input value is None")

            try:
                if node.schema is None:
                    resolved_value = value
                else:
                    resolved_value = node.schema.resolve(value)
            except ValueError:
                raise ValueError(f"Input value resolved failed.")

            if node.on_set is not None:
                await call_mix(node.on_set, resolved_value)

        else:
            raise ValueError(f"Input doesn't support action type: {action_type.value}")



class Interactive:
    def __init__(self) -> None:
        self._interactive_id: str = ""

    @property
    def interactive_id(self) -> str:
        return self._interactive_id

    def set_interactive_id(self, interactive_id: str) -> None:
        self._interactive_id = interactive_id



from acte.node.node import Node

from acte.node.element.element import Element

from acte.node.element.block.block import Block
from acte.node.element.block.root import Root
from acte.node.element.block.div import Div

from acte.node.element.inline.inline import Inline
from acte.node.element.inline.text import Text
from acte.node.element.inline.button import Button
from acte.node.element.inline.input import Input

from acte.node.virtual.virtual import Virtual
from acte.node.virtual.component_node import ComponentNode
from acte.node.virtual.dyna import Dyna
from acte.node.virtual.cache import Cache

__all__ = [
    'Node',
    'Element',
    'Block',
    'Root',
    'Div',
    'Inline',
    'Text',
    'Button',
    'Input',
    'Virtual',
    'ComponentNode',
    'Dyna',
    'Cache',

    'implement',
]



from acte.schema.schema import Schema

from acte.schema.simple_schema.simple_schema import SimpleSchema

from acte.schema.simple_schema.bool_schema import BoolSchema
from acte.schema.simple_schema.int_schema import IntSchema
from acte.schema.simple_schema.num_schema import NumSchema
from acte.schema.simple_schema.str_schema import StrSchema
from acte.schema.simple_schema.object_schema import ObjectSchema
from acte.schema.simple_schema.array_schema import ArraySchema
from acte.schema.simple_schema.null_schema import NullSchema

__all__ = [
    'Schema',
    'SimpleSchema',

    'BoolSchema',
    'IntSchema',
    'NumSchema',
    'StrSchema',
    'ObjectSchema',
    'ArraySchema',
    'NullSchema'
]



tools = [
    {
        "type": "function",
        "function": {
            "name": "new_session",
            "strict": True,
            "description": "Start a new App session, then display the session's latest screen.",
        },
    },
    {
        "type": "function",
        "function": {
            "name": "execute",
            "strict": True,
            "description": "Execute one or more action(s), then display the session's latest screen.",
            "parameters": {
                "type": "object",
                "properties": {
                    "session_id": {
                        "type": "string",
                    },
                    "actions": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "target_id": {
                                    "type": "string",
                                    "description": "Targeted element's id in the view.",
                                },
                                "action_type": {
                                    "type": "string",
                                    "enum": ["press", "set"],
                                    "description": "press: press the button;"
                                                   "set: set the input field's value"
                                },
                                "value": {
                                    "description": "When the action_type is 'input', the value is the input value.",
                                }
                            },
                            "required": ["target_id", "action_type"],
                        },
                    },
                },
                "required": ["session_id", "actions"],
            }
        }
    },
    {
        "type": "function",
        "function": {
            "name": "display",
            "strict": True,
            "description": "Display the session's latest screen.",
            "parameters": {
                "type": "object",
                "properties": {
                    "session_id": {
                        "type": "string",
                    },
                },
                "required": ["session_id"],
            }
        }
    },
]



from __future__ import annotations

import inspect
from typing import Callable, Awaitable, TypeVar, Any

from acte.state.signal import Signal
from acte.context.context import Context

T = TypeVar('T')


class Effect:
    def __init__(self, func: Callable[[], Awaitable[None] | None]) -> None:
        self._func = func

        self._signal_list: list[Signal[Any]] = []
        self._effect_children: list[Effect] = []

        self._is_pending = True
        self._is_cancel = False

        pending_effect_list = Context.get_pending_effect_list()
        pending_effect_list.append(self)

    @classmethod
    async def create(cls, func: Callable[[], Awaitable[None] | None]) -> Effect:
        e = Effect(func)
        await e.async_init()
        return e

    @property
    def is_pending(self) -> bool:
        return self._is_pending

    @property
    def is_cancel(self) -> bool:
        return self._is_cancel

    async def async_init(self) -> None:
        if self._is_pending is False:
            raise ValueError("Effect is not pending")

        async_init_list = Context.get_pending_effect_list()
        async_init_list.remove(self)

        await self.call()

        self._is_pending = False

    async def reset(self) -> None:
        if self._is_cancel is True:
            raise ValueError("Effect is canceled, cannot reset")

        if self._is_pending is True:
            raise ValueError("Effect is pending, cannot reset")

        self.cancel()

        self._is_cancel = False
        self._is_pending = True

        await self.call()

    async def call(self) -> None:
        if self._is_cancel is True:
            raise ValueError("Effect is canceled, cannot call")

        effect_stack = Context.get_effect_stack()

        # add this effect to its parent effect's children
        parent_effect = effect_stack[-1] if len(effect_stack) > 0 else None
        if parent_effect is not None:
            parent_effect._effect_children.append(self)

        # call this effect
        effect_stack.append(self)

        if inspect.iscoroutinefunction(self._func):
            await self._func()
        else:
            self._func()

        effect_stack.pop()

    def add_signal(self, signal: Signal[Any]) -> None:
        self._signal_list.append(signal)

    def remove_signal(self, signal: Signal[Any]) -> None:
        self._signal_list.remove(signal)

    def cancel(self) -> None:
        if self._is_cancel is True:
            raise ValueError("Effect is already canceled")

        # cancel effect children
        for effect in self._effect_children:
            if effect.is_cancel is False:  # parent effect change may have cancelled child effect by its own
                effect.cancel()

        self._effect_children = []

        # remove this effect from all signals
        for signal in self._signal_list:
            signal.remove_effect(self)
            self._signal_list.remove(signal)

        self._is_pending = False
        self._is_cancel = True



from acte.node.element.inline.inline import Inline
from acte.state import Ref, Effect


class Text(Inline):
    def __init__(self) -> None:
        super().__init__()

        self._content: str = ""

    @property
    def content(self) -> str:
        return self._content

    async def bind_content(self, content: Ref[str]) -> None:
        async def _func() -> None:
            v = content.value
            if v is None:
                v = ""

            self._content = v

        effect = await Effect.create(_func)

        self._effect_list.append(effect)



from abc import ABC, abstractmethod
from typing import Any

import jsonschema  # type: ignore

from acte.schema.schema import Schema


class SimpleSchema(Schema, ABC):
    def __init__(self) -> None:
        self._title: str | None = None

    @property
    def title(self) -> str | None:
        return self._title

    def set_title(self, title: str | None) -> None:
        self._title = title

    @abstractmethod
    def resolve(self, data: Any) -> Any:
        pass



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from acte.node.node import Node


class Container:
    def __init__(self) -> None:
        super().__init__()

        self._children: list[Node] = []

    @property
    def children(self) -> list[Node]:
        return self._children

    def set_children(self, children: list[Node]) -> None:
        self._children = children



from acte.node.element.element import Element
from acte.node.implement.container import Container


class Block(Element, Container):
    def __init__(self) -> None:
        Element.__init__(self)
        Container.__init__(self)



from typing import Any


class RetVal:
    def __init__(self) -> None:
        self.session_id: str | None = None
        self.screen: str | None = None

        self.validate_error: dict[str, Any] | None = None
        self.build_error: str | None = None
        self.execute_error: dict[str, Any] | None = None
        self.display_error: str | None = None
        self.executed_actions: list[dict[str, Any]] | None = None

        self.ret_dict: dict[str, Any] = {}

    def to_dict(self) -> dict[str, Any]:
        self.ret_dict = {}

        self._add_dict_field('session_id')
        self._add_dict_field('screen')

        self._add_dict_error_field('build_error')
        self._add_dict_error_field('validate_error')
        self._add_dict_error_field('execute_error')
        self._add_dict_error_field('display_error')
        self._add_dict_error_field('executed_actions')

        return self.ret_dict

    def _add_dict_field(self, field_name: str) -> None:
        value = getattr(self, field_name)
        if value is not None:
            self.ret_dict[field_name] = value

    def _add_dict_error_field(self, field_name: str) -> None:
        value = getattr(self, field_name)
        if value is not None:
            if 'error_info' not in self.ret_dict:
                self.ret_dict['error_info'] = {}

            self.ret_dict['error_info'][field_name] = value



from typing import Callable, TypeVar

from acte.state.ref import Ref

T = TypeVar('T')


class Compute(Ref[T]):
    # noinspection PyMissingConstructor
    def __init__(self, func: Callable[[], T]) -> None:
        self._func = func

    @property
    def value(self) -> T:
        return self._func()



from typing import TypeVar, Type

from acte.build import Component, Builder
from acte.executor import Executor, ActionType
from acte.node import Root, Node
from acte.render import Renderer
from acte.context.context import Context, ContextObj

T = TypeVar('T')


class Session:
    def __init__(
            self,
            entry_class: Type[Component],
            builder: Builder,
            executor: Executor,
            renderer: Renderer
    ) -> None:
        self._entry_class = entry_class
        self._builder = builder
        self._executor = executor
        self._renderer = renderer

        self._context = ContextObj()
        self._root: Root | None = None

    async def start(self) -> None:
        Context.use(self._context)

        self._root = await self._builder.build(self._entry_class)

    async def execute(self, target_id: str, action_type: ActionType, value: str | None = None) -> None:
        Context.use(self._context)

        if not isinstance(self._root, Root):
            raise RuntimeError("session hasn't started, and root is None.")

        await self._executor.execute(self._root, target_id, action_type, value)

    async def display(self) -> None:
        if not isinstance(self._root, Node):
            raise RuntimeError("session hasn't started, and root is None.")

        await self._executor.display(self._root)

    def render(self) -> str:
        if not isinstance(self._root, Node):
            raise RuntimeError("session hasn't started, and root is None.")

        output = self._renderer.render(self._root)

        return output



from acte.state.ref import Ref
from acte.state.signal import Signal
from acte.state.effect import Effect
from acte.state.compute import Compute
from acte.state.memo import Memo

__all__ = [
    'Ref',
    'Signal',
    'Effect',
    'Compute',
    'Memo',
]



import json

import aiohttp
from typing import Any, AsyncGenerator

from acte.chatbot.chatbot import Chatbot
from acte.chatbot.tools import tools


class OpenaiChatbot(Chatbot):
    def __init__(
            self,
            api_key: str,
            system_message: str = '',
            model: str = "gpt-4o",
            temperature: float = 0.3,
            max_tokens: int = 2000,
            base_url: str = "https://api.openai.com/v1",
    ) -> None:
        self._api_key = api_key
        self._system_message = system_message
        self._model = model
        self._temperature = temperature
        self._max_tokens = max_tokens
        self._base_url = base_url

    async def completion(self, messages: list[dict[str, Any]]) -> AsyncGenerator[bytes, None]:
        # Openai doesn't support user tool call and response, so we need to escape them
        messages = [
            {'role': 'system', 'content': self._system_message},
            *self._escape_user_tool_calls_and_responses(messages),
        ]

        if self._system_message != '':
            messages.insert(
                0,
                {'role': 'system', 'content': self._system_message},
            )

        url = f"{self._base_url}/chat/completions"

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self._api_key}"
        }

        payload = {
            "model": self._model,
            "temperature": self._temperature,
            "max_tokens": self._max_tokens,
            "messages": messages,
            "tools": tools,
            "stream": True,
        }

        # use aiohttp
        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=payload, headers=headers) as response:
                if response.content_type == "text/event-stream":
                    async for line in response.content:
                        yield line
                else:
                    resp_str = await response.text()
                    raise Exception(resp_str)

    @staticmethod
    def _escape_user_tool_calls_and_responses(messages: list[dict[str, Any]]) -> list[dict[str, Any]]:
        prev_user_call = False

        new_messages = []

        for message in messages:
            new_message = message

            if 'tool_calls' in message:
                new_message = {
                    "role": "user",
                    "content": json.dumps(message['tool_calls'])
                }
                prev_user_call = True
            elif prev_user_call is True:
                prev_user_call = False
                new_message = {
                    "role": "user",
                    "content": message['content']
                }

            new_messages.append(new_message)

        return new_messages



from __future__ import annotations

from typing import TYPE_CHECKING

from acte.node.virtual.virtual import Virtual
from acte.node.implement.container import Container

if TYPE_CHECKING:
    from acte.build.component import Component


class ComponentNode(Virtual, Container):
    def __init__(self) -> None:
        Virtual.__init__(self)
        Container.__init__(self)

        self._component: Component | None = None

    @property
    def component(self) -> Component | None:
        return self._component

    def set_component(self, component: Component) -> None:
        self._component = component



class SessionValidateException(Exception):
    pass


class SessionBuildException(Exception):
    pass


class SessionExecuteException(Exception):
    pass


class SessionDisplayException(Exception):
    pass



from acte.build.component import Component, call_on_display
from acte.build.viewer.viewer import Viewer
from acte.build.builder import Builder
from acte.build.type import Prop, as_prop

__all__ = [
    "call_on_display",
    "Component",
    "Viewer",
    "Builder",
    "Prop",
    "as_prop",
]



import inspect
from typing import Any, Awaitable, Callable, TypeVar, Union, cast

T = TypeVar('T')


async def call_mix(
        func: Callable[..., Union[Awaitable[T], T]],
        *args: Any,
        **kwargs: Any
) -> T:
    if inspect.iscoroutinefunction(func):
        return cast(T, await func(*args, **kwargs))
    else:
        return cast(T, func(*args, **kwargs))



from __future__ import annotations

from typing import Awaitable, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from acte.node.implement.container import Container
    from acte.node.virtual.cache import Cache
    from acte.state.effect import Effect


class ContextObj:
    def __init__(self) -> None:
        # Used for view construction, component and dyna's view construction has two phases: sync and async
        self._awaitable_list_stack: list[list[Awaitable[None]]] = []

        # collect pending effects in sync phase, then call them in async phase
        self._pending_effect_list: list[Effect] = []

        # in async phase, track containers to ensure children are added to the correct container
        self._container_stack: list[Container] = []

        # in async phase, interactive count is used to define interactive id
        self._interactive_count: int = 0

        # in async phase, track calling effects because effects may be nested
        self._effect_stack: list[Effect] = []

        # skip when cached
        self._is_skip: bool = False

        # save new cache list for each dyna
        self._cache_dict_stack: list[dict[Any, Cache]] = []

    @property
    def awaitable_list_stack(self) -> list[list[Awaitable[None]]]:
        return self._awaitable_list_stack

    def set_awaitable_list_stack(self, awaitable_list_stack: list[list[Awaitable[None]]]) -> None:
        self._awaitable_list_stack = awaitable_list_stack

    @property
    def pending_effect_list(self) -> list[Effect]:
        return self._pending_effect_list

    def set_pending_effect_list(self, pending_effect_list: list[Effect]) -> None:
        self._pending_effect_list = pending_effect_list

    @property
    def container_stack(self) -> list[Container]:
        return self._container_stack

    def set_container_stack(self, container_stack: list[Container]) -> None:
        self._container_stack = container_stack

    @property
    def interactive_count(self) -> int:
        return self._interactive_count

    def set_interactive_count(self, interactive_count: int) -> None:
        self._interactive_count = interactive_count

    @property
    def effect_stack(self) -> list[Effect]:
        return self._effect_stack

    def set_effect_stack(self, effect_stack: list[Effect]) -> None:
        self._effect_stack = effect_stack

    @property
    def is_skip(self) -> bool:
        return self._is_skip

    def set_is_skip(self, is_skip: bool) -> None:
        self._is_skip = is_skip

    @property
    def cache_dict_stack(self) -> list[dict[Any, Cache]]:
        return self._cache_dict_stack

    def set_cache_dict_stack(self, cache_dict_stack: list[dict[Any, Cache]]) -> None:
        self._cache_dict_stack = cache_dict_stack



from typing import Any

import jsonschema  # type: ignore

from acte.schema.simple_schema.simple_schema import SimpleSchema


class StrSchema(SimpleSchema):
    def __init__(self) -> None:
        super().__init__()
        self._enum: list[str] | None = None

    @property
    def enum(self) -> list[str] | None:
        return self._enum

    def set_enum(self, enum: list[str] | None) -> None:
        self._enum = enum

    @property
    def json_schema(self) -> dict[str, Any]:
        schema: dict[str, Any] = {
            "type": "string",
        }

        if self._title is not None:
            schema['title'] = self._title

        if self._enum is not None:
            schema['enum'] = self._enum

        return schema

    def resolve(self, data: Any) -> str:
        jsonschema.validate(data, self.json_schema)
        return str(data)



from abc import ABC, abstractmethod
from typing import Any

import jsonschema  # type: ignore


class Schema(ABC):
    @abstractmethod
    def resolve(self, data: Any) -> Any:
        pass

    @property
    @abstractmethod
    def json_schema(self) -> dict[str, Any]:
        pass



from typing import Callable, Awaitable, cast, Any

from acte.build.type import Prop, to_ref
from acte.build.viewer.common.base import Base

from acte.node import Input
from acte.schema.schema import Schema
from acte.schema.simple_schema.bool_schema import BoolSchema
from acte.schema.simple_schema.int_schema import IntSchema
from acte.schema.simple_schema.num_schema import NumSchema
from acte.schema.simple_schema.str_schema import StrSchema
from acte.state import Signal, Compute


class InputViewer(Base):
    @classmethod
    def input_str(
            cls,
            name: Callable[[], str] | Prop[str] = '',
            value: Prop[str] | None = None,
            on_set: Prop[Callable[[str], Awaitable[None] | None]] | None = None,
            schema: Prop[StrSchema] | None = None,
    ) -> None:
        if schema is None:
            schema = StrSchema()

        cls._input(
            name,
            value,
            on_set,
            schema,
        )

    @classmethod
    def input_int(
            cls,
            name: Callable[[], str] | Prop[str] = '',
            value: Prop[int] | None = None,
            on_set: Prop[Callable[[str], Awaitable[None] | None]] | None = None,
            schema: Prop[IntSchema] | None = None,
    ) -> None:
        if schema is None:
            schema = IntSchema()

        cls._input(
            name,
            value,
            on_set,
            schema,
        )

    @classmethod
    def input_float(
            cls,
            name: Callable[[], str] | Prop[str] = '',
            value: Signal[float] | None = None,
            on_set: Prop[Callable[[str], Awaitable[None] | None]] | None = None,
            schema: Prop[NumSchema] | None = None,
    ) -> None:
        if schema is None:
            schema = NumSchema()

        cls._input(
            name,
            value,
            on_set,
            schema,
        )

    @classmethod
    def input_bool(
            cls,
            name: Callable[[], str] | Prop[str] = '',
            value: Signal[bool] | None = None,
            on_set: Prop[Callable[[str], Awaitable[None] | None]] | None = None,
            schema: Prop[BoolSchema] | None = None,
    ) -> None:
        if schema is None:
            schema = BoolSchema()

        cls._input(
            name,
            value,
            on_set,
            schema,
        )

    @classmethod
    def _input(
            cls,
            name: Callable[[], str] | Prop[str],
            value: Prop[Any] | None,
            on_set: Prop[Callable[[str], Awaitable[None] | None]] | None,
            schema: Prop[Schema],
    ) -> None:
        cls._check_skip()

        if callable(name):
            name = Compute(name)

        if value is None:
            value = Signal(None)

        if (on_set is None) and isinstance(value, Signal):
            value = cast(Signal, value)

            async def on_set(v: Any) -> None:
                await value.set(v)

        cls._append_awaitable(
            cls._input_constructor(
                name,
                value,
                on_set,
                schema,
            )
        )

    @classmethod
    async def _input_constructor(
            cls,
            name: Prop[str],
            value: Prop[Any],
            on_set: Prop[Callable[[str], Awaitable[None] | None]] | None,
            schema: Prop[Schema],
    ) -> None:
        name = to_ref(name)
        value = to_ref(value)
        on_set = to_ref(on_set)
        schema = to_ref(schema)

        node = Input()
        node.set_interactive_id(cls._generate_interactive_id())

        await node.bind_name(name)
        await node.bind_value(value)
        await node.bind_on_set(on_set)
        await node.bind_schema(schema)

        cls._attach_to_container(node)



from typing import Callable, Awaitable, TypeVar, Generic, Type, TypeAlias, Any

from acte.node.element.inline.inline import Inline
from acte.node.implement.interactive import Interactive
from acte.schema.schema import Schema
from acte.state import Ref, Effect, Signal


class Input(Inline, Interactive):
    def __init__(self) -> None:
        Inline.__init__(self)
        Interactive.__init__(self)

        self._name: str = ''
        self._schema: Schema | None = None
        self._value: Any = None
        self._on_set: Callable[[str], Awaitable[None] | None] | None = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def schema(self) -> Schema | None:
        return self._schema

    @property
    def value(self) -> Any:
        return self._value

    @property
    def on_set(self) -> Callable[[str], Awaitable[None] | None] | None:
        return self._on_set

    async def bind_name(self, name: Ref[str]) -> None:
        async def _func() -> None:
            v = name.value
            if v is None:
                v = ""

            self._name = v

        effect = await Effect.create(_func)

        self._effect_list.append(effect)

    async def bind_schema(self, schema: Ref[Schema]) -> None:
        async def _func() -> None:
            self._schema = schema.value

        effect = await Effect.create(_func)

        self._effect_list.append(effect)

    async def bind_value(self, value: Ref[Any]) -> None:
        async def _func() -> None:
            self._value = value.value

        effect = await Effect.create(_func)

        self._effect_list.append(effect)

    async def bind_on_set(self, on_set: Ref[Callable[[str], Awaitable[None] | None]]) -> None:
        async def _func() -> None:
            self._on_set = on_set.value

        effect = await Effect.create(_func)

        self._effect_list.append(effect)



from typing import Callable

from acte.build.viewer.common.base import Base
from acte.node import Dyna


class DynaViewer(Base):
    @classmethod
    def dyna(cls, func: Callable[[], None]) -> None:
        cls._check_skip()

        def decorator(view: Callable[[], None]) -> None:
            cls._append_awaitable(cls._dyna_constructor(view))

        return decorator(func)

    @classmethod
    async def _dyna_constructor(cls, view: Callable[[], None]) -> None:
        node = Dyna()

        async def _on_view_update() -> None:
            cls._append_container(node)

            cls._append_new_awaitable_list()

            cls._append_new_cache_dict()

            view()

            # start: remove children and unmount non-cache children, exclude persistent
            new_cache_dict = cls._pop_new_cache_dict()

            # start: add persistent cache to new_cache_dict
            for identifier, cache in node.cached_dict.items():
                if cache.persistent and (identifier not in new_cache_dict):
                    new_cache_dict[identifier] = cache
            # end: add persistent cache to new_cache_dict

            node.set_cached_dict(new_cache_dict)

            cached_list = list(new_cache_dict.values())

            while len(node.children) != 0:
                child = node.children.pop(0)

                if child not in cached_list:
                    child.unmount(cached_list)
            # end: remove children and unmount non-cache children

            await cls._async_init(node)
            await cls._call_awaitable_list()

            cls._pop_awaitable_list()

            cls._pop_container_node()

        await node.bind_view(view, _on_view_update)

        cls._attach_to_container(node)



from __future__ import annotations

from typing import Awaitable, cast, Any

from acte.build.viewer.common.exception import SkipException
from acte.node import Node, Cache, Dyna
from acte.node.implement import Container
from acte.context.context import Context


class Base:
    @classmethod
    def _append_container(cls, container: Container) -> None:
        container_stack = Context.get_container_stack()
        container_stack.append(container)

    @classmethod
    def _pop_container(cls) -> Container:
        container_stack = Context.get_container_stack()
        return container_stack.pop()

    @classmethod
    def _pop_container_node(cls) -> Node:
        container_stack = Context.get_container_stack()
        container = container_stack.pop()
        return cast(Node, container)

    @classmethod
    def _attach_to_container(cls, node: Node) -> None:
        container_stack = Context.get_container_stack()
        container_stack[-1].children.append(node)

    @classmethod
    def _get_dyna(cls) -> Dyna | None:
        nearest_dyna_node = None
        for container in reversed(Context.get_container_stack()):
            if isinstance(container, Dyna):
                nearest_dyna_node = container
                break

        return nearest_dyna_node

    @classmethod
    def _get_cached(cls, identifier: Any) -> Cache | None:
        dyna_node = cls._get_dyna()
        if dyna_node is not None:
            cache_dict = dyna_node.cached_dict
            return cache_dict.get(identifier)

        return None

    @classmethod
    def _add_cached(cls, cache: Cache) -> None:
        dyna_node = cls._get_dyna()
        if dyna_node is not None:
            cache_dict = dyna_node.cached_dict
            if cache.identifier in cache_dict:
                raise ValueError(f"Duplicated cache identifier: {cache.identifier}")
            cache_dict[cache.identifier] = cache

    @classmethod
    def _add_new_cache(cls, cache: Cache) -> None:
        cache_dict = cls._get_cache_dict()
        cache_dict[cache.identifier] = cache

    @classmethod
    def _append_new_cache_dict(cls) -> None:
        cache_list_stack = Context.get_cache_dict_stack()
        cache_list_stack.append({})

    @classmethod
    def _pop_new_cache_dict(cls) -> dict[Any, Cache]:
        return Context.get_cache_dict_stack().pop()

    @classmethod
    def _get_cache_dict(cls) -> dict[Any, Cache]:
        return Context.get_cache_dict_stack()[-1]

    @classmethod
    def _append_new_awaitable_list(cls) -> None:
        awaitable_list_stack = Context.get_awaitable_list_stack()
        awaitable_list_stack.append([])

    @classmethod
    def _pop_awaitable_list(cls) -> list[Awaitable[None]]:
        awaitable_list_stack = Context.get_awaitable_list_stack()
        return awaitable_list_stack.pop()

    @classmethod
    async def _call_awaitable_list(cls) -> None:
        awaitable_list = Context.get_awaitable_list_stack()[-1]
        for sub in awaitable_list:
            await sub

    @classmethod
    def _append_awaitable(cls, awaitable: Awaitable[None]) -> None:
        awaitable_list = Context.get_awaitable_list_stack()[-1]
        awaitable_list.append(awaitable)

    @classmethod
    def _generate_interactive_id(cls) -> str:
        interactive_id = Context.get_interactive_count()
        Context.set_interactive_count(interactive_id + 1)
        return str(interactive_id)

    @classmethod
    async def _async_init(cls, node: Node) -> None:
        pending_effect_list = Context.get_pending_effect_list()

        for e in pending_effect_list:  # just add top level effect, not children
            node.add_effect(e)

        for e in pending_effect_list:
            await e.async_init()

    @classmethod
    def _set_skip(cls) -> None:
        Context.set_is_skip(True)

    @classmethod
    def _check_skip(cls) -> None:
        if Context.get_is_skip() is True:
            Context.set_is_skip(False)
            raise SkipException()



from json import JSONDecodeError

import uvicorn
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.routing import Route

from acte.chatbot import Chatbot
from acte.server.chatbot_bp import ChatbotBp
from acte.session.session_manager import SessionManager


class Server:
    def __init__(
            self,
            session_manager: SessionManager,
            chatbot: Chatbot | None = None,
    ):
        self._session_manager = session_manager

        self._app = Starlette(routes=[
            Route('/session', self._new_session, methods=['POST']),
            Route('/execute', self._execute, methods=['POST']),
            Route('/display', self._display, methods=['POST']),
        ])

        if chatbot is not None:
            chatbot_bp = ChatbotBp(chatbot)
            self._app.mount("/", chatbot_bp.as_app())

    @property
    def app(self) -> Starlette:
        return self._app

    def run(self, *args, **kwargs) -> None:
        uvicorn.run(self._app, *args, **kwargs)

    async def _new_session(self, request: Request) -> JSONResponse:
        resp_dict = await self._session_manager.new_session()
        return JSONResponse(resp_dict)

    async def _execute(self, request: Request) -> JSONResponse:
        try:
            data = await request.json()
        except JSONDecodeError as e:
            return JSONResponse({"error_info": {"validation_error": str(e)}})

        resp_dict = await self._session_manager.execute(data)
        return JSONResponse(resp_dict)

    async def _display(self, request: Request) -> JSONResponse:
        try:
            data = await request.json()
        except JSONDecodeError as e:
            return JSONResponse({"error_info": {"validation_error": str(e)}})

        resp_dict = await self._session_manager.display(data)
        return JSONResponse(resp_dict)



from __future__ import annotations

from acte.build.component import Component, call_on_mount
from acte.build.viewer.common.base import Base

from acte.node import ComponentNode


class ComponentViewer(Base):
    @classmethod
    def component(cls, c: Component) -> None:
        cls._check_skip()

        cls._append_awaitable(cls._component_constructor(c))

    @classmethod
    async def _component_constructor(cls, c: Component) -> None:
        component_node = ComponentNode()
        component_node.set_component(c)

        cls._append_container(component_node)

        cls._append_new_awaitable_list()

        c.view()
        await call_on_mount(c)
        await cls._async_init(component_node)
        await cls._call_awaitable_list()

        cls._pop_awaitable_list()

        node = cls._pop_container_node()

        cls._attach_to_container(node)



import contextlib
from typing import Iterator

from acte.build.viewer.common.base import Base
from acte.node import Div


class DivViewer(Base):
    @classmethod
    @contextlib.contextmanager
    def div(cls) -> Iterator[None]:
        cls._check_skip()

        cls._append_awaitable(cls._enter_div_constructor())

        try:
            yield
        finally:

            cls._append_awaitable(cls._leave_div_constructor())

    @classmethod
    async def _enter_div_constructor(cls) -> None:
        node = Div()

        cls._append_container(node)

    @classmethod
    async def _leave_div_constructor(cls) -> None:
        node = cls._pop_container_node()
        cls._attach_to_container(node)



from __future__ import annotations

from acte.build.viewer.button_viewer import ButtonViewer
from acte.build.viewer.cache_viewer import CacheViewer
from acte.build.viewer.component_viewer import ComponentViewer
from acte.build.viewer.div_viewer import DivViewer
from acte.build.viewer.dyna_viewer import DynaViewer
from acte.build.viewer.input_viewer import InputViewer
from acte.build.viewer.text_viewer import TextViewer


class Viewer(
    DivViewer,
    TextViewer,
    ButtonViewer,
    InputViewer,

    ComponentViewer,
    DynaViewer,
    CacheViewer,
):
    pass



from __future__ import annotations

from typing import TYPE_CHECKING, Awaitable, Callable

from acte.node.implement.container import Container
from acte.state.effect import Effect

if TYPE_CHECKING:
    from acte.node.virtual.cache import Cache


class Node:
    """
    - Node
        - Element
            - Block (HasChildren)
            - Inline
        - Virtual (HasChildren)
    """

    def __init__(self) -> None:
        self._effect_list: list[Effect] = []

    def add_effect(self, e: Effect) -> None:
        self._effect_list.append(e)

    def unmount(self, cached_list: list[Cache] | None = None) -> None:
        if isinstance(self, Container):
            while len(self.children) != 0:
                child = self.children.pop(0)

                if (cached_list is not None) and (child not in cached_list):
                    child.unmount(cached_list)

        for effect in self._effect_list:
            effect.cancel()



from __future__ import annotations

import inspect
from typing import TypeVar
from abc import abstractmethod

T = TypeVar('T')


async def call_on_mount(c: Component) -> None:
    on_mount_func = getattr(c, 'on_mount', None)
    if on_mount_func is None:
        return

    if inspect.iscoroutinefunction(on_mount_func):
        await on_mount_func()
    else:
        on_mount_func()


async def call_on_display(c: Component) -> None:
    on_display_func = getattr(c, 'on_display', None)
    if on_display_func is None:
        return

    if inspect.iscoroutinefunction(on_display_func):
        await on_display_func()
    else:
        on_display_func()


class Component:
    @abstractmethod
    def view(self) -> None:
        pass



from typing import Callable, Awaitable

from acte.node.element.inline.inline import Inline
from acte.node.implement.interactive import Interactive
from acte.schema.simple_schema.null_schema import NullSchema
from acte.state import Effect, Ref


class Button(Inline, Interactive):
    def __init__(self) -> None:
        Inline.__init__(self)
        Interactive.__init__(self)

        self._content: str = ''
        self._schema: NullSchema | None = None
        self._on_press: Callable[[], Awaitable[None] | None] | None = None

    @property
    def content(self) -> str:
        return self._content

    @property
    def schema(self) -> NullSchema | None:
        return self._schema

    @property
    def on_press(self) -> Callable[[], Awaitable[None] | None] | None:
        return self._on_press

    async def bind_content(self, content: Ref[str]) -> None:
        async def _func() -> None:
            v = content.value
            if v is None:
                v = ''

            self._content = v

        effect = await Effect.create(_func)

        self._effect_list.append(effect)

    async def bind_schema(self, schema: Ref[NullSchema]) -> None:
        async def _func() -> None:
            self._schema = schema.value

        effect = await Effect.create(_func)

        self._effect_list.append(effect)

    async def bind_on_press(self, on_press: Ref[Callable[[], Awaitable[None] | None]]) -> None:
        async def _func() -> None:
            self._on_press = on_press.value

        effect = await Effect.create(_func)

        self._effect_list.append(effect)



from __future__ import annotations

from contextvars import ContextVar
from typing import TYPE_CHECKING, Any, Awaitable

from acte.context.context_obj import ContextObj

if TYPE_CHECKING:
    from acte.node import Cache
    from acte.node.implement import Container
    from acte.state.effect import Effect


class Context:
    _context_obj: ContextVar[ContextObj | None] = ContextVar("context_obj", default=None)

    @classmethod
    def use(cls, c: ContextObj) -> None:
        cls._context_obj.set(c)

    @classmethod
    def get_awaitable_list_stack(cls) -> list[list[Awaitable[None]]]:
        context = cls._context_obj.get()

        if isinstance(context, ContextObj):
            return context.awaitable_list_stack
        else:
            raise ValueError("context_obj is None")

    @classmethod
    def get_pending_effect_list(cls) -> list[Effect]:
        context = cls._context_obj.get()

        if isinstance(context, ContextObj):
            return context.pending_effect_list
        else:
            raise ValueError("context_obj is None")

    @classmethod
    def get_container_stack(cls) -> list[Container]:
        context = cls._context_obj.get()

        if isinstance(context, ContextObj):
            return context.container_stack
        else:
            raise ValueError("context_obj is None")

    @classmethod
    def get_interactive_count(cls) -> int:
        context = cls._context_obj.get()

        if isinstance(context, ContextObj):
            return context.interactive_count
        else:
            raise ValueError("context_obj is None")

    @classmethod
    def set_interactive_count(cls, count: int) -> None:
        context = cls._context_obj.get()

        if isinstance(context, ContextObj):
            context.set_interactive_count(count)
        else:
            raise ValueError("context_obj is None")

    @classmethod
    def get_effect_stack(cls) -> list[Effect]:
        context = cls._context_obj.get()

        if isinstance(context, ContextObj):
            return context.effect_stack
        else:
            raise ValueError("context_obj is None")

    @classmethod
    def get_is_skip(cls) -> bool:
        context = cls._context_obj.get()

        if isinstance(context, ContextObj):
            return context.is_skip
        else:
            raise ValueError("context_obj is None")

    @classmethod
    def set_is_skip(cls, is_skip: bool) -> None:
        context = cls._context_obj.get()

        if isinstance(context, ContextObj):
            context.set_is_skip(is_skip)
        else:
            raise ValueError("context_obj is None")

    @classmethod
    def get_cache_dict_stack(cls) -> list[dict[Any, Cache]]:
        context = cls._context_obj.get()

        if isinstance(context, ContextObj):
            return context.cache_dict_stack
        else:
            raise ValueError("context_obj is None")



import json
import xml.etree.ElementTree as Et

from acte.node import Node, Block, Virtual, Inline, Div, Text, ComponentNode, Root, Button, Dyna, Cache, Input
from acte.node.implement import Container
from acte.render.renderer import Renderer


class HtmlRenderer(Renderer):
    def __init__(self) -> None:
        self._output = ""

    def render(self, node: Node) -> str:
        self._render(node)

        output = self._output

        self._output = ""

        return output

    def _render(self, node: Node) -> None:
        if isinstance(node, Virtual):
            self._render_virtual(node)
        elif isinstance(node, Block):
            self._render_block(node)
        elif isinstance(node, Inline):
            self._render_inline(node)
        else:
            raise ValueError(f"Unknown node type: {type(node)}")

    def _render_virtual(self, virtual: Virtual) -> None:
        if type(virtual) is ComponentNode:
            pass
        elif type(virtual) is Dyna:
            pass
        elif type(virtual) is Cache:
            pass
        else:
            raise ValueError(f"Unknown virtual type: {type(virtual)}")

        if isinstance(virtual, Container):
            for child in virtual.children:
                self._render(child)

        if type(virtual) is ComponentNode:
            pass
        elif type(virtual) is Dyna:
            pass
        elif type(virtual) is Cache:
            pass
        else:
            raise ValueError(f"Unknown virtual type: {type(virtual)}")

    def _render_block(self, block: Block) -> None:
        if type(block) is Root:
            self._output += "<body>"
        elif type(block) is Div:
            self._output += "<div>"
        else:
            raise ValueError(f"Unknown block type: {type(block)}")

        for child in block.children:
            self._render(child)

        if type(block) is Root:
            self._output += "</body>"
        elif type(block) is Div:
            self._output += "</div>"
        else:
            raise ValueError(f"Unknown block type: {type(block)}")

    def _render_inline(self, inline: Node) -> None:
        if isinstance(inline, Text):
            el = Et.Element("span")
            el.text = inline.content

            self._output += Et.tostring(el, method='html', encoding="unicode")
        elif isinstance(inline, Button):
            el = Et.Element(
                "button",
                id=inline.interactive_id,
            )

            if inline.schema is not None:
                el.set("schema", str(inline.schema.json_schema))

            el.text = inline.content

            self._output += Et.tostring(el, method='html', encoding="unicode")
        elif isinstance(inline, Input):
            el = Et.Element("input")

            el.set("id", inline.interactive_id)

            if inline.name != '':
                el.set("name", inline.name)

            if inline.schema is not None:
                el.set("schema", str(inline.schema.json_schema))

            v = json.dumps(inline.value, ensure_ascii=False)
            el.set("value", v)

            self._output += Et.tostring(el, method='html', encoding="unicode")
        else:
            raise ValueError(f"Unknown inline type: {type(inline)}")



from typing import Any

from acte.schema.simple_schema.simple_schema import SimpleSchema


class NullSchema(SimpleSchema):
    @property
    def json_schema(self) -> dict[str, Any]:
        schema: dict[str, Any] = {
            "type": "null",
        }

        if self._title is not None:
            schema['title'] = self._title

        return schema

    def resolve(self, data: Any) -> None:
        return None



from abc import ABC, abstractmethod

from typing import AsyncGenerator, Any


class Chatbot(ABC):
    @abstractmethod
    def completion(self, messages: list[dict[str, Any]]) -> AsyncGenerator[bytes, None]:
        pass



from typing import Any

from acte.session import Session


class Validator:
    @classmethod
    def validate_execute_request(
            cls,
            data: dict[str, Any],
            session_dict: dict[str, Session],
    ) -> dict[str, Any] | None:
        err_dict: dict[str, Any] = {}

        sid_err = cls._validate_session_id(data, session_dict)
        if sid_err is not None:
            err_dict.update(sid_err)

        actions = data.get('actions', None)
        if actions is None:
            err_dict['actions'] = 'Missing field.'
        else:
            if not isinstance(actions, list):
                err_dict['actions'] = 'Invalid value. Must be a list.'
            else:
                action_err_info_list: list[Any] = []
                for i, action in enumerate(actions):
                    action_err_info = cls._validate_action(action)
                    if action_err_info is not None:
                        item = {"index": i, **action_err_info}
                        action_err_info_list.append(item)

                if len(action_err_info_list) != 0:
                    err_dict['actions'] = action_err_info_list

        if len(err_dict) != 0:
            return err_dict

        return None

    @classmethod
    def _validate_action(cls, action_data: dict[str, Any]) -> dict[str, Any] | None:
        err_info_dict: dict[str, Any] = {}

        if 'target_id' not in action_data:
            err_info_dict['target_id'] = 'Missing field.'
        else:
            if not isinstance(action_data['target_id'], str):
                err_info_dict['target_id'] = 'Invalid value. Must be a string.'

        if 'action_type' not in action_data:
            err_info_dict['action_type'] = 'Missing field.'
        else:
            if action_data['action_type'] not in ['press', 'set']:
                err_info_dict['action_type'] = 'Invalid value. Must be one of "press" and "set". '

            if action_data['action_type'] == 'set':
                if 'value' not in action_data:
                    err_info_dict['value'] = 'Missing field.'
            else:
                if 'value' in action_data:
                    err_info_dict['value'] = 'Invalid field. Must not be present.'

        if len(err_info_dict) != 0:
            return err_info_dict

        return None

    @classmethod
    def _validate_session_id(cls, data: dict[str, Any], session_dict: dict[str, Session]) -> dict[str, Any] | None:
        err_dict: dict[str, Any] = {}

        sid = data.get('session_id', None)
        if sid is None:
            err_dict['session_id'] = 'Missing field.'
        else:
            if not isinstance(sid, str):
                err_dict['session_id'] = 'Invalid value. Must be a string.'
            else:
                if sid not in session_dict:
                    err_dict['session_id'] = 'Invalid value. Session not found.'

        if len(err_dict) != 0:
            return err_dict

        return None



from __future__ import annotations

from typing import Callable, Awaitable, Any, TYPE_CHECKING

from acte.node.virtual.virtual import Virtual
from acte.node.implement.container import Container
from acte.state import Effect

if TYPE_CHECKING:
    from acte.node import Cache


class Dyna(Virtual, Container):
    def __init__(self) -> None:
        Virtual.__init__(self)
        Container.__init__(self)

        self._view: Callable[[], None] = lambda: None

        self._cached_dict: dict[Any, Cache] = {}

    @property
    def view(self) -> Callable[[], None]:
        return self._view

    @property
    def cached_dict(self) -> dict[Any, Cache]:
        return self._cached_dict

    def set_cached_dict(self, value: dict[Any, Cache]) -> None:
        self._cached_dict = value

    async def bind_view(self, view: Callable[[], None], on_view_update: Callable[[], Awaitable[None]]) -> None:
        self._view = view

        async def _func() -> None:
            await on_view_update()

        e = await Effect.create(_func)
        self.add_effect(e)



from typing import Callable, Awaitable, cast

from acte.build.type import Prop, to_ref
from acte.build.viewer.common.base import Base

from acte.node import Button
from acte.schema.simple_schema.null_schema import NullSchema
from acte.state import Compute, Ref, Signal, Effect


class ButtonViewer(Base):
    @classmethod
    def button(
            cls,
            content: Callable[[], str] | Prop[str] = '',
            on_press: Prop[Callable[[], Awaitable[None] | None]] | None = None,
            schema: Prop[NullSchema] | None = None,
    ) -> None:
        cls._check_skip()

        if callable(content):
            content = Compute(content)

        if schema is None:
            schema = NullSchema()

        cls._append_awaitable(
            cls._button_constructor(
                content,
                schema,
                on_press,
            )
        )

    @classmethod
    async def _button_constructor(
            cls,
            content: Prop[str],
            schema: Prop[NullSchema],
            on_press: Prop[Callable[[], Awaitable[None] | None]] | None,
    ) -> None:
        content = to_ref(content)
        schema = to_ref(schema)
        on_press = to_ref(on_press)

        node = Button()
        node.set_interactive_id(cls._generate_interactive_id())

        await node.bind_content(content)
        await node.bind_schema(schema)

        if on_press is not None:
            await node.bind_on_press(on_press)

        cls._attach_to_container(node)



from typing import Type

from acte.build.viewer.viewer import Viewer
from acte.build.component import Component
from acte.node import Root


class Builder(Viewer):
    @classmethod
    async def build(cls, entry_class: Type[Component]) -> Root:
        entry = entry_class()

        root = Root()

        cls._append_container(root)

        await cls._component_constructor(entry)

        cls._pop_container()

        return root



import inspect
from typing import Callable, TypeVar, cast, Awaitable

from acte.state.effect import Effect
from acte.state.signal import Signal
from acte.state.ref import Ref

T = TypeVar('T')


class Memo(Ref[T]):
    # noinspection PyMissingConstructor
    def __init__(self, func: Callable[[], Awaitable[T] | T]) -> None:
        self._func = func
        self._signal: Signal[T] | None = None

        async def _effect_func() -> None:
            if inspect.iscoroutinefunction(self._func):
                value = await self._func()
            else:
                value = self._func()

            if self._signal is None:
                self._signal = Signal(value)
            else:
                await self._signal.set(value)

        self._effect = Effect(_effect_func)

    @property
    def is_pending(self) -> bool:
        return self._effect.is_pending

    @property
    def effect(self) -> Effect:
        return self._effect

    @property
    def value(self) -> T | None:
        signal = cast(Signal[T], self._signal)

        return signal.value

    async def async_init(self) -> None:
        if self._effect.is_pending is False:
            raise ValueError("Memo is already async initialized")

        await self._effect.async_init()

    def cancel(self) -> None:
        if self._effect.is_cancel is True:
            raise ValueError("Memo is already canceled")

        self._effect.cancel()



from __future__ import annotations

import contextlib
from typing import Any, Iterator, cast, Container

from acte.build.viewer.common.exception import SkipException
from acte.build.viewer.common.base import Base

from acte.node import Cache


class CacheViewer(Base):
    @classmethod
    @contextlib.contextmanager
    def cache(cls, identifier: Any, persistent: bool = False) -> Iterator[None]:
        cls._check_skip()

        cached = cls._get_cached(identifier)

        if cached is None:
            cls._append_awaitable(cls._enter_new_cache_constructor(identifier, persistent))
        else:
            cls._append_awaitable(cls._enter_cached_constructor(cached))

            cls._set_skip()
            cls._add_new_cache(cached)

        try:
            yield
        except SkipException:
            pass

        finally:
            cls._append_awaitable(cls._leave_cache_constructor())

    @classmethod
    async def _enter_new_cache_constructor(cls, identifier: Any, persistent: bool) -> None:
        cache = Cache(identifier, persistent)
        cls._add_cached(cache)  # execute during await cls._call_awaitable_list(), so add to cached

        cls._append_container(cache)

    @classmethod
    async def _enter_cached_constructor(cls, cached: Cache) -> None:
        cls._append_container(cached)

    @classmethod
    async def _leave_cache_constructor(cls) -> None:
        node = cls._pop_container_node()
        cls._attach_to_container(node)



import os
import typing
from json import JSONDecodeError

from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import JSONResponse, StreamingResponse
from starlette.routing import Route, Mount, BaseRoute
from starlette.staticfiles import StaticFiles

from acte.chatbot import Chatbot


class _CustomStaticFiles(StaticFiles):
    def file_response(self, full_path, stat_result, scope, status_code=200):
        response = super().file_response(full_path, stat_result, scope, status_code)
        if full_path.endswith('.js'):
            response.headers['Content-Type'] = 'application/javascript'
        return response


class ChatbotBp:
    def __init__(self, chatbot: Chatbot, load_static: bool = True) -> None:
        self._chatbot = chatbot
        self._load_static = load_static

    def as_app(self) -> Starlette:
        routes: list[BaseRoute] = [Route('/completions', self._completions_route, methods=['POST'])]

        if self._load_static:
            this_file_dir = os.path.dirname(os.path.abspath(__file__))
            static_dir = os.path.join(this_file_dir, "static")

            static_route = Mount('/', app=_CustomStaticFiles(directory=static_dir, html=True), name="static")
            routes.append(static_route)

        return Starlette(routes=routes)

    async def _completions_route(self, request: Request) -> JSONResponse | StreamingResponse:
        try:
            request_dict = await request.json()
        except JSONDecodeError:
            return JSONResponse({"error_info": {"validation_error": "fail to decode json"}}, 400)

        try:
            messages = request_dict["messages"]
        except KeyError:
            return JSONResponse({"err_info": {"validation_error": "messages key not found"}}, 400)

        generator = self._chatbot.completion(messages)

        try:
            first_result = await anext(generator)
        except Exception as e:
            return JSONResponse({"error_info": {"validation_error": str(e)}}, 400)

        async def generate():
            yield first_result

            async for choice in generator:
                yield choice

        return StreamingResponse(generate(), media_type="text/event-stream")



from typing import Generic, TypeVar

T = TypeVar('T', covariant=True)


class Ref(Generic[T]):
    def __init__(self, value: T|None) -> None:
        self._value = value

    @property
    def value(self) -> T|None:
        return self._value



from typing import Any

from acte.node.virtual.virtual import Virtual
from acte.node.implement.container import Container


class Cache(Virtual, Container):
    def __init__(self, identifier: Any, persistent: bool) -> None:
        Virtual.__init__(self)
        Container.__init__(self)

        self._identifier = identifier
        self._persistent = persistent

    @property
    def identifier(self) -> Any:
        return self._identifier

    @property
    def persistent(self) -> bool:
        return self._persistent
